package dev.fitko.fitconnect.integrationtests;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.nimbusds.jose.jwk.RSAKey;
import dev.fitko.fitconnect.api.config.ApplicationConfig;
import dev.fitko.fitconnect.api.config.SchemaConfig;
import dev.fitko.fitconnect.api.domain.model.event.Event;
import dev.fitko.fitconnect.api.domain.model.event.SubmissionStatus;
import dev.fitko.fitconnect.api.domain.model.metadata.ContentStructure;
import dev.fitko.fitconnect.api.domain.model.metadata.Hash;
import dev.fitko.fitconnect.api.domain.model.metadata.Metadata;
import dev.fitko.fitconnect.api.domain.model.metadata.SignatureType;
import dev.fitko.fitconnect.api.domain.model.metadata.attachment.ApiAttachment;
import dev.fitko.fitconnect.api.domain.model.metadata.attachment.Purpose;
import dev.fitko.fitconnect.api.domain.model.metadata.data.Data;
import dev.fitko.fitconnect.api.domain.model.metadata.data.MimeType;
import dev.fitko.fitconnect.api.domain.model.metadata.data.SubmissionSchema;
import dev.fitko.fitconnect.api.domain.model.replychannel.ReplyChannel;
import dev.fitko.fitconnect.api.exceptions.client.FitConnectSenderException;
import dev.fitko.fitconnect.api.services.crypto.CryptoService;
import dev.fitko.fitconnect.client.SenderClient;
import dev.fitko.fitconnect.client.SubscriberClient;
import dev.fitko.fitconnect.client.factory.ClientFactory;
import dev.fitko.fitconnect.client.sender.model.Attachment;
import dev.fitko.fitconnect.client.sender.model.SendableEncryptedSubmission;
import dev.fitko.fitconnect.client.sender.model.SendableSubmission;
import dev.fitko.fitconnect.client.subscriber.ReceivedSubmission;
import dev.fitko.fitconnect.core.crypto.HashService;
import dev.fitko.fitconnect.core.crypto.JWECryptoService;
import dev.fitko.fitconnect.integrationtests.condition.EnableIfEnvironmentVariablesAreSet;
import org.apache.tika.mime.MimeTypes;
import org.awaitility.Awaitility;
import org.hamcrest.MatcherAssert;
import org.hamcrest.Matchers;
import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;

import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.text.ParseException;
import java.time.Duration;
import java.util.List;
import java.util.UUID;

import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.containsString;
import static org.hamcrest.Matchers.hasSize;
import static org.hamcrest.Matchers.is;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertThrows;

@EnableIfEnvironmentVariablesAreSet
public class SenderClientIT extends IntegrationTestBase {

    @BeforeEach
    public void cleanup(){
        cleanupTestSubmissions();
    }

    @Test
    void testSendAndConfirmCycle() throws IOException {

        // Given
        final ApplicationConfig config = getConfigWithCredentialsFromEnvironment("TESTING", true);

        final String submissionData = getResourceAsString("/submission_data.json");

        final var submission = SendableSubmission.Builder()
                .setDestination(UUID.fromString(System.getenv("TEST_DESTINATION_ID")))
                .setServiceType("urn:de:fim:leika:leistung:99400048079000", "Test Service")
                .setJsonData(submissionData, URI.create("https://schema.fitko.de/fim/s00000114_1.1.schema.json"))
                .addAttachment(Attachment.fromPath(Path.of("src/test/resources/attachment.txt"), "text/plain"))
                .addAttachment(Attachment.fromByteArray("attachment data".getBytes(), "text/plain"))
                .addAttachment(Attachment.fromString("attachment data", "text/plain"))
                .setReplyChannel(ReplyChannel.fromDeMail("test@mail.org"))
                .build();

        final var sentSubmission = ClientFactory.getSenderClient(config).send(submission);

        assertNotNull(sentSubmission);

        // When
        final ReceivedSubmission receivedSubmission =
                ClientFactory.getSubscriberClient(config)
                        .requestSubmission(sentSubmission.getSubmissionId());

        // Then
        assertNotNull(receivedSubmission);
        assertThat(receivedSubmission.getDataAsString(), is(submissionData));
        assertThat(receivedSubmission.getDataSchemaUri(), is(URI.create("https://schema.fitko.de/fim/s00000114_1.1.schema.json")));
        assertThat(receivedSubmission.getDataMimeType(), is("application/json"));
        assertThat(receivedSubmission.getAttachments(), hasSize(3));
        assertThat(receivedSubmission.getMetadata().getReplyChannel(), is(ReplyChannel.fromDeMail("test@mail.org")));
        assertThat(new String(receivedSubmission.getAttachments().get(0).getDataAsBytes()), is("Test attachment"));
    }

    @Test
    void testSendAndConfirmCycleWithEncryptedData() throws ParseException, IOException {

        // Given
        final ApplicationConfig config = getConfigWithCredentialsFromEnvironment("TESTING", true);
        final UUID destinationId = UUID.fromString(System.getenv("TEST_DESTINATION_ID"));

        final CryptoService cryptoService = new JWECryptoService(new HashService());
        final SenderClient senderClient = ClientFactory.getSenderClient(config);
        final SubscriberClient subscriberClient = ClientFactory.getSubscriberClient(config);

        final RSAKey encryptionKey = RSAKey.parse(senderClient.getPublicKeyForDestination(destinationId));

        final String jsonData = getResourceAsString("/submission_data.json");
        final File attachmentFile = new File("src/test/resources/attachment.txt");
        final String attachmentData = Files.readString(attachmentFile.toPath());

        final String encryptedData = cryptoService.encryptBytes(encryptionKey, jsonData.getBytes(StandardCharsets.UTF_8));
        final String encryptedAttachment = cryptoService.encryptBytes(encryptionKey, attachmentData.getBytes(StandardCharsets.UTF_8));

        final var submissionSchema = new SubmissionSchema();
        submissionSchema.setSchemaUri(URI.create("https://schema.fitko.de/fim/s00000114_1.1.schema.json"));
        submissionSchema.setMimeType(MimeType.APPLICATION_JSON);

        final var dataHash = new Hash();
        dataHash.setSignatureType(SignatureType.SHA_512);
        dataHash.setContent(cryptoService.hashBytes(jsonData.getBytes(StandardCharsets.UTF_8)));

        final var data = new Data();
        data.setHash(dataHash);
        data.setSubmissionSchema(submissionSchema);

        final var attachment = new ApiAttachment();
        attachment.setAttachmentId(UUID.randomUUID());
        attachment.setPurpose(Purpose.ATTACHMENT);
        attachment.setFilename(attachmentFile.getName());
        attachment.setMimeType(MimeTypes.PLAIN_TEXT);

        final var attachmentHash = new Hash();
        attachmentHash.setContent(cryptoService.hashBytes(attachmentData.getBytes(StandardCharsets.UTF_8)));
        attachmentHash.setSignatureType(SignatureType.SHA_512);
        attachment.setHash(attachmentHash);

        final var contentStructure = new ContentStructure();
        contentStructure.setAttachments(List.of(attachment));
        contentStructure.setData(data);

        final var metadata = new Metadata();
        metadata.setSchema(SchemaConfig.METADATA_V_1_0_0.toString());
        metadata.setContentStructure(contentStructure);

        final String encryptedMetadata = cryptoService.encryptBytes(encryptionKey, new ObjectMapper().writeValueAsBytes(metadata));

        // When
        final var submission = SendableEncryptedSubmission.Builder()
                .setDestination(destinationId)
                .setServiceType("urn:de:fim:leika:leistung:99400048079000", "Test Service")
                .setEncryptedMetadata(encryptedMetadata)
                .setEncryptedData(encryptedData)
                .addEncryptedAttachment(attachment.getAttachmentId(), encryptedAttachment)
                .build();

        final var sentSubmission = senderClient.send(submission);

        Assertions.assertNotNull(sentSubmission);

        final ReceivedSubmission receivedSubmission = subscriberClient.requestSubmission(sentSubmission.getSubmissionId());

        // Then
        Assertions.assertNotNull(receivedSubmission);
        MatcherAssert.assertThat(receivedSubmission.getDataAsString(), Matchers.is(jsonData));
        MatcherAssert.assertThat(receivedSubmission.getDataMimeType(), Matchers.is("application/json"));
        MatcherAssert.assertThat(receivedSubmission.getAttachments().get(0).getDataAString(StandardCharsets.UTF_8), Matchers.is("Test attachment"));
    }

    @Test
    void testAbortedSendSubmissionWithKeyValidationNotSilent() throws IOException {

        // Given
        final ApplicationConfig config = getConfigWithCredentialsFromEnvironment("PROD", false);

        // When
        final var submission = SendableSubmission.Builder()
                .setDestination(UUID.fromString(System.getenv("TEST_DESTINATION_ID")))
                .setServiceType("urn:de:fim:leika:leistung:99400048079000", "Test Service")
                .setJsonData(getResourceAsString("/submission_data.json"), URI.create("https://schema.fitko.de/fim/s00000114_1.1.schema.json"))
                .build();

        final FitConnectSenderException exception = assertThrows(FitConnectSenderException.class, () -> ClientFactory.getSenderClient(config).send(submission));

        // Then
        assertThat(exception.getMessage(), containsString("Sending submission failed"));
        assertThat(exception.getCause().getMessage(), containsString("JWK with id jsMQHFiA2uMGtiG3Ro8RJnYdEhp5W5KjGW-Vcf3-YMk has an invalid amount of certificates. Is 1 but should be 3."));
    }

    @Test
    void testReadSubmissionStatus() throws IOException {

        // Given
        final ApplicationConfig config = getConfigWithCredentialsFromEnvironment("TESTING", true);

        final String submissionData = getResourceAsString("/submission_data.json");

        final SenderClient senderClient = ClientFactory.getSenderClient(config);

        final var submission = SendableSubmission.Builder()
                .setDestination(UUID.fromString(System.getenv("TEST_DESTINATION_ID")))
                .setServiceType("urn:de:fim:leika:leistung:99400048079000", "Test Service")
                .setJsonData(submissionData, URI.create("https://schema.fitko.de/fim/s00000114_1.1.schema.json"))
                .addAttachment(Attachment.fromPath(Path.of("src/test/resources/attachment.txt"), "plain/text"))
                .build();

        final var sentSubmission = senderClient.send(submission);

        Assertions.assertNotNull(sentSubmission);

        Awaitility.await().atMost(Duration.ofSeconds(30)).until(() ->
        {
            // When
            final SubmissionStatus statusForSubmission = senderClient.getStatusForSubmission(sentSubmission);

            // Then
            return statusForSubmission.getStatus().equals(Event.SUBMIT.getState());
        });
    }

}
